// ****************************************************************************
//             Copyright (C) Oasys Design Systems, Inc. 2004 - 2008
//                             All rights reserved.
// ****************************************************************************
#include "LLoad.h"
#include "attr_enum.h"
#include "group_enum.h"
#include "NCell"
#include "NMVCell"
#include "NRef"
#include "NDb"
#include "TLutTemplate.h"
#include "LibLoadMsg.h"
#include "LibParam.h"

void
LLoad::addCell(const LibGroup* lcell)
{
  NCell cell, tmpCell;

  uassert(lcell->firstName() && !lcell->moreThanOneName());
  if (UParam::libertyLoadVerbosity()) {
    LIBLOADMSG(loadingCell, (str_t)lcell->firstName());
  }

#if 1
  bool isMvCell = false;
  preProcessCellAttr(lcell, isMvCell);
  if (!isMvCell) {
    preProcessCellSubgroup(lcell, isMvCell);
  }
#endif

  if (UParam::libAllowMultiLibsForCell()) {
    cell = d_lib.database().findCell(lcell->firstName());
    if (!cell) {
      tmpCell = d_lib.newCell(lcell->firstName());
    } else {
      UStrBuf cellName;
      cellName.format_string("%s");
      cellName << library().name() << ":" << lcell->firstName();
      tmpCell = d_lib.newCell(cellName.str());
    }
    if (isMvCell) {
      (void) tmpCell.getMVCell();
    }

    // process cell attributes first
    LLoadRailList railList; 
    processCellAttr(lcell, tmpCell, railList);
    processCellSubgroup(lcell, tmpCell);
    instantiateCellPins(lcell, tmpCell, railList);

    if (cell) {
      // now merge their corner sensitive info
      if (cell.recharacterize(tmpCell)) {
	d_lib.removeCell(tmpCell);
	uassert(!tmpCell.hasLibertyInfo());
	d_lib.database().delCell(tmpCell); // also remove the object
	d_lib.addCell(cell);
      } else {
	LIBLOADMSG(cellAlreadyLoaded, (str_t)lcell->firstName());

	// set the libertyInfo so that all layered data are also deleted
	tmpCell.setHasLibertyInfo(true);
	d_lib.removeCell(tmpCell);
	d_lib.database().delCell(tmpCell); // also remove the object
      }
    } else {
      cell = tmpCell;
      cell.characterize();
      // set the liberty info flag only when everything is done.
      cell.setHasLibertyInfo(true);
    }
  } else {
    if (d_lib.database().findCell(lcell->firstName())) {
      LIBLOADMSG(cellAlreadyLoaded, (str_t)lcell->firstName());
      return;
    }
    cell = d_lib.newCell(lcell->firstName());

    // process cell attributes first
    LLoadRailList railList;
    processCellAttr(lcell, cell, railList);
    processCellSubgroup(lcell, cell);
    instantiateCellPins(lcell, cell, railList);
    cell.setHasLibertyInfo(true);
  }


}

void
LLoad::addLutTemplate(const LibGroup* ltemp)
{
  TLib& tlib = *d_lib.tlib();
  TLutTemplate* temp;
  LibErrorT err;
  LibAttr* attr;
  const char* idxStr[3];
  uint_t i;
  int idxVarType[3];
  UKeyword variable[3] = {K_variable_1, K_variable_2, K_variable_3};
  UKeyword index[3] = {K_index_1, K_index_2, K_index_3};

  const UString& lutName = ltemp->firstName();

  uassert(lutName && !ltemp->moreThanOneName());

  for (i=0; i<3; i++) {
    attr = ltemp->findAttr(variable[i], err);
    if (!attr) {
      break;
    }
    idxVarType[i] = UString(attr->value()->string_val()).keyword();
    attr = ltemp->findAttr(index[i], err);
    if (!attr) {
      LIBLOADMSG(missingLutTemplateIndex, lutName.str(), i+1, i+1);
      return;
    }
    idxStr[i] = attr->firstValue()->string_val();
  }
  if (!i) {
    LIBLOADMSG(missingLutTemplateVariable, lutName.str());
    return;
  }
  temp = tlib.findLutTemplate(lutName);
  if (temp) {
    LIBLOADMSG(lutTemplateAlreadyLoaded, lutName.str());
    return;
  }
  temp = tlib.addTemplate(lutName);
  temp->templateIndexTable(i, idxStr, idxVarType, tlib);
}

void
LLoad::processScalingFactorsGroup(const LibGroup* sfGroup)
{
  LibAttrIter aiter;
  LibAttr*    attr;
  TLib&       tlib = *d_lib.tlib();
  const UString& sfGroupName = sfGroup->firstName();

  TLibScaleFactors* sf = tlib.findScaleFactors(sfGroupName);
  if (sf) {
    LIBLOADMSG(scaleFactorsAlreadyLoaded, sfGroupName.str());
    return;
  }
  sf = tlib.addScaleFactors(sfGroupName);
  aiter = sfGroup->attrIter();
  while ((attr = aiter++)) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_k_process_cell_fall                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_cell_leakage_power           : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell_leakage_power(TLKFD_process, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_cell_rise                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_drive_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_drive_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_fall_propagation             : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_fall_transition              : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_hold_fall                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_hold_rise                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_internal_power               : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_internal_power(TLKFD_process, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_intrinsic_fall               : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_intrinsic_rise               : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_pin_cap                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_pin_cap(TLKFD_process, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_recovery_fall                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_recovery_rise                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_removal_fall                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_removal_rise                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_rise_propagation             : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_rise_transition              : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_setup_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_setup_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_slope_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_process, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_slope_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_process, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_cell_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_cell_leakage_power              : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell_leakage_power(TLKFD_temperature, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_cell_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_drive_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_drive_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_fall_propagation                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_fall_transition                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_hold_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_hold_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_internal_power                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_internal_power(TLKFD_temperature, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_intrinsic_fall                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_intrinsic_rise                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_pin_cap                         : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_pin_cap(TLKFD_temperature, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_recovery_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_recovery_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_removal_fall                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_removal_rise                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_rise_propagation                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_rise_transition                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_setup_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_setup_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_slope_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_temperature, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_temp_slope_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_temperature, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_cell_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_cell_leakage_power              : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell_leakage_power(TLKFD_voltage, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_cell_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_cell(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_drive_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_drive_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_drive(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_fall_propagation                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_fall_transition                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_hold_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_hold_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_hold(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_internal_power                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_internal_power(TLKFD_voltage, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_intrinsic_fall                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_intrinsic_rise                  : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_intrinsic(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_pin_cap                         : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_pin_cap(TLKFD_voltage, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_recovery_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_recovery_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_recovery(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_removal_fall                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_removal_rise                    : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_removal(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_rise_propagation                : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_propagation(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_rise_transition                 : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_transition(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_setup_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_setup_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_setup(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_slope_fall                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_voltage, TEdge::fall, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_volt_slope_rise                      : // float( > -100.0 and  < 100.0 ) := 0.0;
      sf->k_slope(TLKFD_voltage, TEdge::rise, attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_k_process_cell_degradation             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_drive_current                : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_fall_delay_intercept         : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_fall_pin_resistance          : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_min_period                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_min_pulse_width_high         : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_min_pulse_width_low          : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_nochange_fall                : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_nochange_rise                : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_rise_delay_intercept         : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_rise_pin_resistance          : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_skew_fall                    : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_skew_rise                    : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_wire_cap                     : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_process_wire_res                     : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_cell_degradation                : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_drive_current                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_fall_delay_intercept            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_fall_pin_resistance             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_min_period                      : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_min_pulse_width_high            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_min_pulse_width_low             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_nochange_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_nochange_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_rise_delay_intercept            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_rise_pin_resistance             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_skew_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_skew_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_wire_cap                        : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_temp_wire_res                        : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_cell_degradation                : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_drive_current                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_fall_delay_intercept            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_fall_pin_resistance             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_min_period                      : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_min_pulse_width_high            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_min_pulse_width_low             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_nochange_fall                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_nochange_rise                   : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_rise_delay_intercept            : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_rise_pin_resistance             : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_skew_fall                       : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_skew_rise                       : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_wire_cap                        : // float( > -100.0 and  < 100.0 ) := 0.0;
    case LIBERTY_ATTRENUM_k_volt_wire_res                        : // float( > -100.0 and  < 100.0 ) := 0.0;
    default:
      if (warnNoSupport()) {
	printf("Attribute %s of scaling_factors group %s on line %d of library file %s"
	       " not supported.\n", attr->name().str(), sfGroupName.str(),
	       sfGroup->lineno(), sfGroup->filename());
      }
      break;
    }
  }
}

void
LLoad::processOperatingCndsGroup(const LibGroup* ocGroup)
{
  LibAttrIter aiter;
  LibAttr*    attr;
  const UString& ocGroupName = ocGroup->firstName();

  TLibOpCnd* oc = d_lib.tlib()->findOpCnd(ocGroupName);
  if (oc) {
    LIBLOADMSG(operatingConditionAlreadyLoaded, ocGroupName.str());
    return;
  }
  oc = d_lib.tlib()->addOpCnd(ocGroupName);
  aiter = ocGroup->attrIter();
  while ((attr = aiter++)) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_process                            : // float( > 0.0 and  < 100.0 ) := 1.0;
      oc->process(attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_temperature                        : // float( > -300.0 and  < 500.0 ) := 25.0;
      oc->temperature(attr->value()->double_val());
      break;
    case LIBERTY_ATTRENUM_voltage                            : // float;
      oc->voltage(attr->value()->double_val());
      break;
    default:
      if (warnNoSupport()) {
	printf("Attribute %s of operating_condition group %s on line %d of library file %s"
	       " not supported.\n", attr->name().str(), ocGroupName.str(),
	       ocGroup->lineno(), ocGroup->filename());
      }
      break;
    }
  }
}

void
LLoad::processPowerSupplyGroup(const LibGroup* psGroup)
{
  LibAttrIter aiter;
  LibAttr*    attr;
  TLib& tlib = *d_lib.tlib();

  const UString& psGroupName = psGroup->firstName();

  aiter = psGroup->attrIter();
  while ((attr = aiter++)) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_power_rail                            : // (string, float)
      {
	LibValIter viter = attr->valueIter();
	LibVal* v = viter++;
	const char* supplyName = v->string_val();
	v = viter++;
	double voltage = v->double_val();
	tlib.addVoltageMap(supplyName, voltage);
      }
      break;
    case LIBERTY_ATTRENUM_default_power_rail                    : // string
      {
	const char* defRail = attr->value()->string_val();
	tlib.setDefaultPowerRail(defRail);
      }
      break;
    default:
      if (warnNoSupport()) {
	printf("Attribute %s of power_supply group %s on line %d of library file %s"
	       " not supported.\n", attr->name().str(), psGroupName.str(),
	       psGroup->lineno(), psGroup->filename());
      }
      break;
    }
  }
}

UAggType*
LLoad::processTypeGroup(const LibGroup* tGroup)
{
  LibAttrIter aiter;
  LibAttr*    attr;

  int from = -1, to = -1, width = -1;
  int downto = -1;
  aiter = tGroup->attrIter();
  while ((attr = aiter++)) {
    switch (attr->typeID()) {
    case LIBERTY_ATTRENUM_base_type :
      uassert (!strcmp(attr->value()->string_val(), "array"));
      break;
    case LIBERTY_ATTRENUM_bit_from  :
      from = (int)attr->value()->double_val();
      break;
    case LIBERTY_ATTRENUM_bit_to    :
      to   = (int)attr->value()->double_val();
      break;
    case LIBERTY_ATTRENUM_bit_width :
      width = (int)attr->value()->double_val();
      break;
    case LIBERTY_ATTRENUM_data_type :
      uassert (!strcmp(attr->value()->string_val(), "bit"));
      break;
    case LIBERTY_ATTRENUM_downto    :
      downto = attr->value()->bool_val();
      break;
    default:
      uassert(0);
      break;
    }
  }
  URange r;
  if (from > -1) {
    r.msb(from);
    if (to > -1) {
      r.lsb(to);
      if (width > -1) {
	uassert ((uint_t)width == r.size());
      }
      uassert ((r.size() <= 1) || (downto < 0) || ((bool)downto == r.downto()));
    } else {
      uassert (width > -1);
      if (downto) {
	to = from - width + 1;
      } else {
	to = from + width - 1;
      }
      uassert (to > -1);
    }
  } else {
    uassert (to > -1);
    uassert (width > -1);
    uassert (downto >= 0);
    if (downto) {
      from = to + width - 1;
    } else {
      from = to - width + 1;
    }
    uassert (from > -1);
  }
  return UAggTypeManager::arrayType(r);
}

void
LLoad::createTypeForLib(const LibGroup* tGroup)
{
  LibAttrIter aiter;
  const UString& typeName = tGroup->firstName();
  TLib* tlib = d_lib.tlib();

  if (tlib->findType(typeName)) {
    LIBLOADMSG(typeAlreadyLoaded, typeName.str());
    return;
  }
  
  UAggType *type = processTypeGroup(tGroup);
  tlib->quickAddType(typeName, type);
}

void
LLoad::processLibSubgroup(const LibGroup* llib, NDb db)
{
  LibGroupIter giter;
  LibGroup* group;
  const char* libName = llib->firstName();

  giter = llib->groupIter();
  while (!db.control().interrupted() && (group = giter++)) {
    switch (group->typeID()) {
    case LIBERTY_GROUPENUM_cell                        : // ( cell_name )
    case LIBERTY_GROUPENUM_scaled_cell                 : // ( scaled_cell_name_1,scaled_cell_name_2 )
      // process the cells later
      break;
    case LIBERTY_GROUPENUM_lu_table_template           : // ( lu_table_template_name )
      addLutTemplate(group);
      break;
    case LIBERTY_GROUPENUM_scaling_factors             : // ( scaling_factors_name )
      processScalingFactorsGroup(group);
      break;
    case LIBERTY_GROUPENUM_operating_conditions        : // ( operating_conditions_name )
      processOperatingCndsGroup(group);
      break;
    case LIBERTY_GROUPENUM_type                        : // ( type_name )
      createTypeForLib(group);
      break;
    case LIBERTY_GROUPENUM_power_supply                : // ( [power_supply_name] )
      processPowerSupplyGroup(group);
      break;

    case LIBERTY_GROUPENUM_poly_template               : // ( poly_template_name )
    case LIBERTY_GROUPENUM_em_lut_template             : // ( em_lut_template_name )
    case LIBERTY_GROUPENUM_fall_net_delay              : // ( fall_net_delay_name )
    case LIBERTY_GROUPENUM_fall_transition_degradation : // ( fall_transition_degradation_name )
    case LIBERTY_GROUPENUM_input_voltage               : // ( input_voltage_name )
    case LIBERTY_GROUPENUM_iv_lut_template             : // ( iv_lut_template_name )
    case LIBERTY_GROUPENUM_model                       : // ( model_name )
    case LIBERTY_GROUPENUM_noise_lut_template          : // ( noise_lut_template_name )
    case LIBERTY_GROUPENUM_output_voltage              : // ( output_voltage_name )
    case LIBERTY_GROUPENUM_power_lut_template          : // ( power_lut_template_name )
    case LIBERTY_GROUPENUM_power_poly_template         : // ( power_poly_template_name )
    case LIBERTY_GROUPENUM_propagation_lut_template    : // ( propagation_lut_template_name )
    case LIBERTY_GROUPENUM_rise_net_delay              : // ( rise_net_delay_name )
    case LIBERTY_GROUPENUM_rise_transition_degradation : // ( rise_transition_degradation_name )
    case LIBERTY_GROUPENUM_timing_range                : // ( timing_range_name )
    case LIBERTY_GROUPENUM_wire_load                   : // ( wire_load_name )
    case LIBERTY_GROUPENUM_wire_load_selection         : // ( [wire_load_selection_name] )
    case LIBERTY_GROUPENUM_wire_load_table             : // ( wire_load_table_name )
    default:
      if (warnNoSupport()) {
	printf("Group %s (in library %s) on line %d of library file %s"
	       " not supported.\n", group->type().str(), libName,
	       group->lineno(), group->filename());
      }
      break;
    }
  }
  // set the operating conditions
  TLib* tlib = d_lib.tlib();
  if (!tlib->defaultOpCndName().empty()) {
    tlib->defaultOpCnd(tlib->defaultOpCndName());
  }
  // set the min/max operating conditions to the default operating condition
  // if no default was specified, it will the nom (typical) specified in
  // the library
  for (TLibCorner::enumerator p; p.ok(); p++) {
    tlib->opCnd(p.value(), tlib->defaultOpCnd());
  }
  // process the cells
  giter = llib->groupIter();
  while (!db.control().interrupted() && (group = giter++)) {
    switch (group->typeID()) {
    case LIBERTY_GROUPENUM_cell                        : // ( cell_name )
      addCell(group);
      break;
    case LIBERTY_GROUPENUM_scaled_cell                 : // ( scaled_cell_name_1,scaled_cell_name_2 )
      if (warnNoSupport()) {
	printf("Group %s (in library %s) on line %d of library file %s"
	       " not supported.\n", group->type().str(), libName,
	       group->lineno(), group->filename());
      }
      break;
    default:
      break;
    }
  }
}
